import pytest

import numpy as np
from scipy import signal

from sosfilt import sosfiltfilt


class TestFiltFilt(object):
    def filtfilt(self, zpk, x, axis=-1, padtype="odd", padlen=None):
        sos = signal.zpk2sos(*zpk)
        return sosfiltfilt(sos, x, axis, padtype, padlen)

    def test_basic(self):
        zpk = signal.tf2zpk([1, 2, 3], [1, 2, 3])
        out = self.filtfilt(zpk, np.arange(12))
        np.testing.assert_allclose(out, np.arange(12), atol=5.28e-11)

    def test_sine(self):
        rate = 2000
        t = np.linspace(0, 1.0, rate + 1)
        xlow = np.sin(5 * 2 * np.pi * t)
        xhigh = np.sin(250 * 2 * np.pi * t)
        x = xlow + xhigh

        zpk = signal.butter(8, 0.125, output="zpk")
        # magnitude of the largest pole
        r = np.abs(zpk[1]).max()
        eps = 1e-5
        # estimate of number of steps to decay
        # transient by factor of eps
        n = int(np.ceil(np.log(eps) / np.log(r)))

        y = self.filtfilt(zpk, x, padlen=n)
        err = np.abs(y - xlow).max()
        np.testing.assert_(err < 1e-4)

        x2d = np.vstack([xlow, xlow + xhigh])
        y2d = self.filtfilt(zpk, x2d, padlen=n, axis=1)
        np.testing.assert_equal(y2d.shape, x2d.shape)
        err = np.abs(y2d - xlow).max()
        np.testing.assert_(err < 1e-4)

        # test axis keyword
        y2dt = self.filtfilt(zpk, x2d.T, padlen=n, axis=0)
        np.testing.assert_equal(y2d, y2dt.T)

    def test_axis(self):
        x = np.arange(10.0 * 11.0 * 12.0).reshape(10, 11, 12)
        zpk = signal.butter(3, 0.125, output="zpk")
        y0 = self.filtfilt(zpk, x, padlen=0, axis=0)
        y1 = self.filtfilt(zpk, np.swapaxes(x, 0, 1), padlen=0, axis=1)
        np.testing.assert_array_equal(y0, np.swapaxes(y1, 0, 1))
        y2 = self.filtfilt(zpk, np.swapaxes(x, 0, 2), padlen=0, axis=2)
        np.testing.assert_array_equal(y0, np.swapaxes(y2, 0, 2))

    def test_equivalence(self):
        x = np.random.RandomState(0).randn(1000)
        for order in range(1, 6):
            zpk = signal.butter(order, 0.35, output="zpk")
            b, a = signal.zpk2tf(*zpk)
            sos = signal.zpk2sos(*zpk)
            y = signal.filtfilt(b, a, x)
            y_sos = sosfiltfilt(sos, x)
            np.testing.assert_allclose(y, y_sos, atol=1e-12, err_msg=f"order={order}")


def test_multiple():
    x = np.random.RandomState(0).randn(2, 20, 2)
    a, b = x.shape[0], x.shape[2]

    filters = np.stack(
        [
            signal.butter(3, np.random.RandomState(0).rand(), output="sos")
            for _ in range(a * b)
        ]
    )

    y_sos = sosfiltfilt(filters, x, axis=1)

    y = np.empty(x.shape, dtype=x.dtype)

    for i in range(a):
        for j in range(b):
            y[i, :, j] = signal.sosfiltfilt(filters[i * b + j, ...], x[i, :, j])

    np.testing.assert_allclose(y, y_sos)



import numpy as np
from scipy import signal

from ._zpk_funcs import zpk2sos_multiple


def _relative_degree(z, p):
    if z.ndim == 1 and p.ndim == 1:
        return len(p) - len(z)
    elif z.ndim == 2 and p.ndim == 2:
        return p.shape[0] - z.shape[0]

    raise ValueError("relative degree between singular and multiple filter")


def lp2lp_zpk(z, p, k, wo=1.0):
    z = np.atleast_1d(z)
    p = np.atleast_1d(p)
    wo = np.atleast_1d(wo)

    degree = _relative_degree(z, p)

    z_lp = wo * z[:, None]
    p_lp = wo * p[:, None]
    k_lp = k * wo ** degree

    return z_lp, p_lp, k_lp


def lp2hp_zpk(z, p, k, wo=1.0):
    z = np.atleast_1d(z)
    p = np.atleast_1d(p)
    wo = np.atleast_1d(wo)

    nfilt = len(wo)

    degree = _relative_degree(z, p)

    z_hp = wo / z[:, None]
    p_hp = wo / p[:, None]

    z_hp = np.append(z_hp, np.zeros((degree, nfilt)), axis=0)

    # gain change is applied to the original z, p args, which
    # are only 1d -- so we can just repeat the result
    k_hp = k * np.real(np.prod(-z) / np.prod(-p))
    k_hp = np.repeat(k_hp, nfilt)

    return z_hp, p_hp, k_hp


def bilinear_zpk_multiple(z, p, k, fs):
    z = np.atleast_2d(z)
    p = np.atleast_2d(p)
    k = np.atleast_1d(k)

    nfilt = len(k)

    degree = _relative_degree(z, p)
    fs2 = 2.0 * fs

    # bilinear transform poles and zeros
    z_z = (fs2 + z) / (fs2 - z)
    p_z = (fs2 + p) / (fs2 - p)

    # zeros at infinity are moved to Nyquist frequency
    z_z = np.append(z_z, -np.ones((degree, nfilt)), axis=0)

    # compensate for gain change
    k_z = k * np.real(np.prod(fs2 - z, axis=0) / np.prod(fs2 - p, axis=0))

    return z_z, p_z, k_z


def zpk2tf_multiple(z, p, k):
    raise NotImplementedError("zpk2tf not implemented yet")


def butter(N, Wn, btype="lowpass", output="sos", fs=None):
    """
    A modified filter design that will return a suite of
    Butterworth filters, for all the provided Wn values.
    """

    try:
        btype = signal.filter_design.band_dict[btype]
    except KeyError as e:
        raise ValueError("'{}' is an invalid bandtype.".format(btype)) from e

    Wn = np.asarray(Wn)
    if fs is not None:
        Wn = 2 * Wn / fs

    # generate the single Nth-order filter at 1 rad/s cutoff
    z, p, k = signal.buttap(N)

    fs = 2.0
    warped = 2 * fs * np.tan(np.pi * Wn / fs)

    if btype == "lowpass":
        z, p, k = lp2lp_zpk(z, p, k, wo=warped)
    elif btype == "highpass":
        z, p, k = lp2hp_zpk(z, p, k, wo=warped)
    else:
        raise NotImplementedError("'{}' is not a supported bandtype".format(btype))

    # after this point, z, p, k are arrays of multiple filters:
    # for z, p: axis 0 is the "normal" shape; axis 1 is filter number
    # k is an array with length number of filters

    z, p, k = bilinear_zpk_multiple(z, p, k, fs=fs)

    if output == "zpk":
        return z, p, k
    elif output == "ba":
        return zpk2tf_multiple(z, p, k)
    elif output == "sos":
        return zpk2sos_multiple(z.astype(np.complex_), p.astype(np.complex_), k)



cimport numpy as np
cimport cython

ctypedef fused DTYPE_t:
    float
    float complex
    double
    double complex
    long double
    long double complex

@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def _sosfilt(DTYPE_t [:, :, ::1] sos,
             DTYPE_t [:, ::1] x,
             DTYPE_t [:, :, ::1] zi):
    cdef Py_ssize_t n_signals = x.shape[0]
    cdef Py_ssize_t n_samples = x.shape[1]
    cdef Py_ssize_t n_sections = sos.shape[1]
    cdef Py_ssize_t i, n, s
    cdef DTYPE_t x_new, x_cur
    cdef DTYPE_t[:, ::1] zi_slice

    with nogil:
        for i in range(n_signals):
            zi_slice = zi[i, :, :]

            for n in range(n_samples):
                x_cur = x[i, n]

                for s in range(n_sections):
                    x_new = sos[i, s, 0] * x_cur + zi_slice[s, 0]
                    zi_slice[s, 0] = sos[i, s, 1] * x_cur - sos[i, s, 4] * x_new + zi_slice[s, 1]
                    zi_slice[s, 1] = sos[i, s, 2] * x_cur - sos[i, s, 5] * x_new
                    x_cur = x_new

                x[i, n] = x_cur



import pytest

import numpy as np

from sosfilt import _zpk_funcs


class TestCplxReal:
    def test_trivial_input(self):
        np.testing.assert_equal(
            _zpk_funcs._cplxreal(np.array([], dtype=np.complex_)), ([], [])
        )
        np.testing.assert_equal(
            _zpk_funcs._cplxreal(np.array([1], dtype=np.complex_)), ([], [1])
        )

    def test_output_order(self):
        zc, zr = _zpk_funcs._cplxreal(np.roots(np.array([1, 0, 0, 1])))
        np.testing.assert_allclose(
            np.append(zc, zr), [1 / 2 + 1j * np.sin(np.pi / 3), -1]
        )

        eps = np.spacing(1)

        a = [
            0 + 1j,
            0 - 1j,
            eps + 1j,
            eps - 1j,
            -eps + 1j,
            -eps - 1j,
            1,
            4,
            2,
            3,
            0,
            0,
            2 + 3j,
            2 - 3j,
            1 - eps + 1j,
            1 + 2j,
            1 - 2j,
            1 + eps - 1j,  # sorts out of order
            3 + 1j,
            3 + 1j,
            3 + 1j,
            3 - 1j,
            3 - 1j,
            3 - 1j,
            2 - 3j,
            2 + 3j,
        ]
        zc, zr = _zpk_funcs._cplxreal(np.array(a))
        np.testing.assert_allclose(
            zc, [1j, 1j, 1j, 1 + 1j, 1 + 2j, 2 + 3j, 2 + 3j, 3 + 1j, 3 + 1j, 3 + 1j]
        )
        np.testing.assert_allclose(zr, [0, 0, 1, 2, 3, 4])

        z = np.array(
            [
                1 - eps + 1j,
                1 + 2j,
                1 - 2j,
                1 + eps - 1j,
                1 + eps + 3j,
                1 - 2 * eps - 3j,
                0 + 1j,
                0 - 1j,
                2 + 4j,
                2 - 4j,
                2 + 3j,
                2 - 3j,
                3 + 7j,
                3 - 7j,
                4 - eps + 1j,
                4 + eps - 2j,
                4 - 1j,
                4 - eps + 2j,
            ]
        )

        zc, zr = _zpk_funcs._cplxreal(z)
        np.testing.assert_allclose(
            zc, [1j, 1 + 1j, 1 + 2j, 1 + 3j, 2 + 3j, 2 + 4j, 3 + 7j, 4 + 1j, 4 + 2j]
        )
        np.testing.assert_equal(zr, [])

    def test_unmatched_conjugates(self):
        with pytest.raises(ValueError):
            _zpk_funcs._cplxreal(np.array([1 + 3j, 1 - 3j, 1 + 2j]))

        with pytest.raises(ValueError):
            _zpk_funcs._cplxreal(np.array([1 + 3j]))

        with pytest.raises(ValueError):
            _zpk_funcs._cplxreal(np.array([1 - 3j]))


class TestZpk2Sos:
    def sos2zpk(self, z, p, k):
        return _zpk_funcs.zpk2sos_multiple(
            z[:, None], p[:, None], np.array([k], dtype=np.float64)
        )[0, ...]

    def test_basic(self):
        z = np.array([-1.0, -1.0], dtype=np.complex_)
        p = np.array([0.57149 + 0.29360j, 0.57149 - 0.29360j])
        k = 1
        sos = self.sos2zpk(z, p, k)
        sos2 = [[1, 2, 1, 1, -1.14298, 0.41280]]
        np.testing.assert_array_almost_equal(sos, sos2, decimal=4)

        z = np.array([1j, -1j])
        p = np.array([0.9, -0.9, 0.7j, -0.7j])
        k = 1
        sos = self.sos2zpk(z, p, k)
        sos2 = [[1, 0, 1, 1, 0, +0.49], [1, 0, 0, 1, 0, -0.81]]
        np.testing.assert_array_almost_equal(sos, sos2, decimal=4)

        z = np.array([], dtype=np.complex_)
        p = np.array([0.8, -0.5 + 0.25j, -0.5 - 0.25j])
        k = 1
        sos = self.sos2zpk(z, p, k)
        sos2 = [[1.0, 0.0, 0.0, 1.0, 1.0, 0.3125], [1.0, 0.0, 0.0, 1.0, -0.8, 0.0]]
        np.testing.assert_array_almost_equal(sos, sos2, decimal=4)

        z = np.array([1, 1, 0.9j, -0.9j])
        p = np.array([0.99 + 0.01j, 0.99 - 0.01j, 0.1 + 0.9j, 0.1 - 0.9j])
        k = 1
        sos = self.sos2zpk(z, p, k)
        sos2 = [[1, 0, 0.81, 1, -0.2, 0.82], [1, -2, 1, 1, -1.98, 0.9802]]
        np.testing.assert_array_almost_equal(sos, sos2, decimal=4)

        z = np.array([0.9 + 0.1j, 0.9 - 0.1j, -0.9])
        p = np.array([0.75 + 0.25j, 0.75 - 0.25j, 0.9])
        k = 1
        sos = self.sos2zpk(z, p, k)
        sos2 = [[1, 0.9, 0, 1, -1.5, 0.625], [1, -1.8, 0.82, 1, -0.9, 0]]
        np.testing.assert_array_almost_equal(sos, sos2, decimal=4)



from setuptools import Extension, setup
from pythran.dist import PythranExtension
import numpy

setup(
    ext_modules=[
        Extension("sosfilt._sosfilt", ["sosfilt/_sosfilt.pyx"]),
        PythranExtension(
            "sosfilt._zpk_funcs", ["sosfilt/_zpk_funcs.py"],
            library_dirs=["/usr/lib64/atlas", "/usr/lib/atlas"],
        ),
    ],
    include_dirs=[numpy.get_include(), "/usr/include"],
)



import numpy as np

# pythran export _cplxreal(complex[])
def _cplxreal(z):
    tol = 100 * np.finfo((1.0 * z).dtype).eps

    z = z[np.lexsort((abs(z.imag), z.real))]
    real_indices = abs(z.imag) <= tol * abs(z)
    zr = z[real_indices].real

    if len(zr) == len(z):
        return np.array([]), zr

    z = z[~real_indices]
    zp = z[z.imag > 0]
    zn = z[z.imag < 0]

    if len(zp) != len(zn):
        raise ValueError("Array contains complex value without conjugate")

    same_real = np.diff(zp.real) <= tol * abs(zp[:-1])
    diffs = np.diff(np.concatenate(([0], same_real, [0])))
    run_starts = np.nonzero(diffs > 0)[0]
    run_stops = np.nonzero(diffs < 0)[0]

    for i in range(len(run_starts)):
        start = run_starts[i]
        stop = run_stops[i] + 1
        for chunk in (zp[start:stop], zn[start:stop]):
            chunk[:] = chunk[np.lexsort([abs(chunk.imag)])]

    zc = (zp + zn.conj()) / 2
    return zc, zr


def _nearest_real_complex_idx(fro, to, which):
    order = np.argsort(np.abs(fro - to))
    mask = np.isreal(fro[order])
    if which == "complex":
        mask = ~mask
    return order[np.nonzero(mask)[0][0]]


def poly(zeros):
    dt = zeros.dtype
    a = np.ones((1,), dtype=dt)
    for k in range(len(zeros)):
        a = np.convolve(a, np.array([1, -zeros[k]], dtype=dt), mode="full")

    return a


def zpk2tf(z, p, k):
    b = k * poly(z)
    a = poly(p)

    # this is a bit of an assumption to make...
    return b.real, a.real


def zpk2sos(z, p, k, n_sections):
    sos = np.zeros((n_sections, 6))

    if len(p) % 2 == 1:
        p = np.append(p, 0)
        z = np.append(z, 0)

    z = np.concatenate(_cplxreal(z))
    p = np.concatenate(_cplxreal(p))

    p_sos = np.zeros((n_sections, 2), np.complex128)
    z_sos = np.zeros_like(p_sos)
    for si in range(n_sections):
        # select the next "worst" pole
        p1_idx = np.argmin(np.abs(1 - np.abs(p)))
        p1 = p[p1_idx]
        p = np.delete(p, p1_idx)

        # pair that pole with a zero

        if np.isreal(p1) and np.isreal(p).sum() == 0:
            # first order section
            z1_idx = _nearest_real_complex_idx(z, p1, "real")
            z1 = z[z1_idx]
            z = np.delete(z, z1_idx)
            p2 = z2 = 0
        else:
            if not np.isreal(p1) and np.isreal(z).sum() == 1:
                # choose a complex zero to pair with
                z1_idx = _nearest_real_complex_idx(z, p1, "complex")
                assert not np.isreal(z[z1_idx])
            else:
                z1_idx = np.argmin(np.abs(p1 - z))

            z1 = z[z1_idx]
            z = np.delete(z, z1_idx)

            # we have p1 and z1, figure out p2 and z2
            if not np.isreal(p1):
                if not np.isreal(z1):  # complex pole, complex zero
                    p2 = p1.conj()
                    z2 = z1.conj()
                else:  # complex pole, real zero
                    p2 = p1.conj()
                    z2_idx = _nearest_real_complex_idx(z, p1, "real")
                    z2 = z[z2_idx]
                    assert np.isreal(z2)
                    z = np.delete(z, z2_idx)
            else:
                if not np.isreal(z1):  # real pole, complex zero
                    z2 = z1.conj()
                    p2_idx = _nearest_real_complex_idx(p, z1, "real")
                    p2 = p[p2_idx]
                    assert np.isreal(p2)
                else:  # real pole, real zero
                    # next "worst" pole
                    idx = np.nonzero(np.isreal(p))[0]
                    assert len(idx) > 0
                    p2_idx = idx[np.argmin(np.abs(np.abs(p[idx]) - 1))]
                    p2 = p[p2_idx]
                    assert np.isreal(p2)
                    z2_idx = _nearest_real_complex_idx(z, p2, "real")
                    z2 = z[z2_idx]
                    assert np.isreal(z2)
                    z = np.delete(z, z2_idx)
                p = np.delete(p, p2_idx)

        p_sos[si] = [p1, p2]
        z_sos[si] = [z1, z2]

    assert len(p) == len(z) == 0
    del p, z

    p_sos = p_sos[::-1]
    z_sos = z_sos[::-1]
    gains = np.ones(n_sections, np.array(k).dtype)
    gains[0] = k
    for si in range(n_sections):
        x = zpk2tf(z_sos[si], p_sos[si], gains[si])
        sos[si] = np.concatenate(x)

    return sos


# pythran export zpk2sos_multiple(complex[][], complex[][], float[])
def zpk2sos_multiple(z, p, k):
    nfilt = len(k)
    assert z.shape[1] == p.shape[1] == nfilt

    # pad p and z to the same length
    p = np.concatenate((p, np.zeros((max(z.shape[0] - p.shape[0], 0), nfilt))), axis=0)
    z = np.concatenate((z, np.zeros((max(p.shape[0] - z.shape[0], 0), nfilt))), axis=0)

    n_sections = (max(p.shape[0], z.shape[0]) + 1) // 2
    sos = np.zeros((nfilt, n_sections, 6))

    for filt in range(nfilt):
        sos[filt, :, :] = zpk2sos(z[:, filt], p[:, filt], k[filt], n_sections)

    return sos



import numpy as np
from scipy.signal import signaltools
from ._sosfilt import _sosfilt


def _validate_sos(sos):
    sos = np.atleast_2d(sos)

    if sos.ndim == 2:
        return sos, -1, 1

    n_filters, n_sections, m = sos.shape

    if m != 6:
        raise ValueError("last dim of sos array must be size 6")
    if not (sos[..., 3] == 1).all():
        raise ValueError("sos[..., 3] should be all ones")

    return sos, n_sections, n_filters


def _validate_nfilters(x, axis, n_filters):
    # with 1 filter, we'll broadcast it as usual
    if n_filters == 1:
        return

    # otherwise, the number of filters must match the
    # product of the non-filtered axes of x
    x_shape = list(x.shape)
    x_shape.pop(axis)
    if n_filters != np.product(x_shape):
        raise ValueError("n_filters must match product of non-filtered axes")


def _validate_ntaps(ntaps):
    first_tap = ntaps[0]
    if not np.all(ntaps == first_tap):
        raise ValueError("all filters must have the same number of taps")

    return first_tap


def sosfilt_zi(sos):
    sos = np.asarray(sos)
    if sos.ndim != 3 or sos.shape[2] != 6:
        raise ValueError("sos must be shape (n_filters, n_sections, 6)")

    if sos.dtype.kind in "bui":
        sos = sos.astype(np.float64)

    n_filters, n_sections = sos.shape[:2]
    zi = np.empty((n_filters, n_sections, 2), dtype=sos.dtype)
    scale = np.ones(n_filters, dtype=sos.dtype)
    for section in range(n_sections):
        b = sos[:, section, :3]
        a = sos[:, section, 3:]

        # lfilter_zi solves zi = A*zi + B
        # where A = scipy.linalg.companion(a).T
        # and   B = b[1:] - a[1:]*b[0]
        #
        # because a[0] = 1 for a sos filter, we have
        # A = [ 1 + a[1], -1;
        #         a[2],    1]
        # so zi[0] = (B[0] + B[1]) / (1 + a[1] + a[2])
        # and zi[1] = B[1] - a[2] * zi[0]
        #
        # we can pretty easily write this in a vectorised
        # way over all n_filters!

        B = b[:, 1:] - a[:, 1:] * b[:, [0]]
        zi[:, section, 0] = B.sum(axis=-1) / a.sum(axis=-1)
        zi[:, section, 1] = B[:, 1] - a[:, 2] * zi[:, section, 0]
        zi[:, section, :] *= scale[:, None]

        scale *= b.sum(axis=-1) / a.sum(axis=-1)

    return zi


def sosfilt(sos, x, axis=-1, zi=None):
    x = signaltools._validate_x(x)

    sos, n_sections, n_filters = _validate_sos(sos)

    x_zi_shape = list(x.shape)
    x_zi_shape[axis] = 2
    x_zi_shape = tuple([n_sections] + x_zi_shape)
    inputs = [sos, x]

    if zi is not None:
        inputs.append(np.asarray(zi))

    dtype = np.result_type(*inputs)
    if dtype.char not in "fdgFDGO":
        raise NotImplementedError(f"input type '{dtype}' not supported")

    if zi is not None:
        zi = np.array(zi, dtype)
        if zi.shape != x_zi_shape:
            raise ValueError("invalid zi shape")
        return_zi = True
    else:
        zi = np.zeros(x_zi_shape, dtype=dtype)
        return_zi = False

    axis = axis % x.ndim
    x = np.moveaxis(x, axis, -1)
    zi = np.moveaxis(zi, [0, axis + 1], [-2, -1])
    x_shape, zi_shape = x.shape, zi.shape
    x = np.reshape(x, (-1, x.shape[-1]))
    x = np.array(x, dtype, order="C")  # make a modifiable copy
    zi = np.ascontiguousarray(np.reshape(zi, (-1, n_sections, 2)))
    sos = sos.astype(dtype, copy=False)

    _sosfilt(sos, x, zi)

    x.shape = x_shape
    x = np.moveaxis(x, -1, axis)

    if return_zi:
        zi.shape = zi_shape
        zi = np.moveaxis(zi, [-2, -1], [0, axis + 1])
        out = (x, zi)
    else:
        out = x

    return out


def sosfiltfilt(sos, x, axis=-1, padtype="odd", padlen=None):
    """
    A forward-backward digital filter using cascaded second-order sections.

    Parameters
    ----------
    sos : array_like
        An array of second-order filter coefficients. It must have either
        the shape ``(n_filters, n_sections, 6)`` or ``(n_sections, 6)``.
        In the latter case, the single filter will be broadcast over the
        whole input array. In the former case, `n_filters` must match
        the product of the non-filter axes of `x`. Additionally,
        `n_sections` must be the same for all filters.
    x : array_like
        The array of data to be filtered.
    axis : int, optional
        The axis of `x` to which the filter is applied.
        Default is -1.
    padtype : str or None, optional
        Must be 'odd', 'even', 'constant', or None. Determines the type
        of extension to use for the padded signal to which the filter is
        applied. If None, no padding is used. The default is 'odd'.
    padlen : int or None, optional
        The number of elements by which to extend `x` at both ends of
        `axis` before applying the filter.
    """

    # filter validation
    sos, n_sections, n_filters = _validate_sos(sos)

    if n_filters == 1:
        # defer to scipy's implementation for the usual case of a single filter
        return signaltools.sosfiltfilt(
            sos, x, axis=axis, padtype=padtype, padlen=padlen
        )

    # input validation
    x = signaltools._validate_x(x)
    _validate_nfilters(x, axis, n_filters)

    # padding validation
    ntaps = np.ones(n_filters, dtype=int) * (2 * n_sections + 1)
    ntaps -= np.minimum((sos[..., 2] == 0).sum(axis=1), (sos[..., 5] == 0).sum(axis=1))
    ntaps = _validate_ntaps(ntaps)
    edge, ext = signaltools._validate_pad(padtype, padlen, x, axis, ntaps=ntaps)

    # filter initial conditions
    zi = sosfilt_zi(sos)

    # to handle multiple filters, we might want zi.shape = (n_sections, ..., 2, ...)
    zi = np.swapaxes(zi, 0, 1)  # => (n_sections, n_filters, 2)
    zi_shape = list(x.shape)
    zi_shape[axis] = 2
    # we need to swap axis to the end first, so it picks up the end of zi
    zi_shape[axis], zi_shape[-1] = zi_shape[-1], zi_shape[axis]
    zi_shape = [n_sections] + zi_shape
    zi = zi.reshape(zi_shape)  # should look like (n_sections, ..., 2)
    # now we need to swap axis back from the end to where it should be
    zi = np.swapaxes(zi, -1, (axis % x.ndim) + 1)

    # forward filter
    x_0 = signaltools.axis_slice(ext, stop=1, axis=axis)
    (y, zf) = sosfilt(sos, ext, axis=axis, zi=zi * x_0)

    # backward filter
    y_0 = signaltools.axis_slice(y, start=-1, axis=axis)
    (y, zf) = sosfilt(
        sos, signaltools.axis_reverse(y, axis=axis), axis=axis, zi=zi * y_0
    )

    # reshaping
    y = signaltools.axis_reverse(y, axis=axis)
    if edge > 0:
        y = signaltools.axis_slice(y, start=edge, stop=-edge, axis=axis)

    return y
